package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionPayload;

import java.net.URI;
import java.util.Objects;
import java.util.UUID;
import java.util.regex.Pattern;

public class ValidDataGuard {

    private static final Pattern LEIKA_KEY_PATTERN = Pattern.compile("^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$");
    private final Sender sender;

    public ValidDataGuard(final Sender sender) {
        this.sender = sender;
    }

    /**
     * Checks if the encrypted submission payload is in a valid state for sending.
     *
     * @param encryptedSubmissionPayload payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     * @throws IllegalStateException    if the destination does not have a declared service type
     */
    public void ensureValidDataPayload(final EncryptedSubmissionPayload encryptedSubmissionPayload) {
        if (encryptedSubmissionPayload == null) {
            throw new IllegalArgumentException("Encrypted payload must not be null.");
        }
        if (encryptedSubmissionPayload.getEncryptedData() == null) {
            throw new IllegalArgumentException("Encrypted data is mandatory, but was null.");
        }
        if (encryptedSubmissionPayload.getEncryptedMetadata() == null) {
            throw new IllegalArgumentException("Encrypted metadata must not be null.");
        }
        testDefaults(encryptedSubmissionPayload.getDestinationId(), encryptedSubmissionPayload.getServiceType());
    }

    /**
     * Checks if the unencrypted submission payload is in a valid state for sending.
     *
     * @param submissionPayload payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     */
    public void ensureValidDataPayload(final SubmissionPayload submissionPayload) {
        if (submissionPayload == null) {
            throw new IllegalArgumentException("Payload must not be null.");
        }
        if (submissionPayload.getData() == null) {
            throw new IllegalArgumentException("Data is mandatory, but was null.");
        }
        testOnValidDataFormat(submissionPayload);
        testDefaults(submissionPayload);
    }

    private void testDefaults(final UUID destinationId, final ServiceType serviceType) {

        if (destinationId == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (serviceType == null) {
            throw new IllegalArgumentException("ServiceType is mandatory, but was null.");
        } else if (serviceType.getIdentifier() == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(serviceType.getIdentifier())) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        Destination destination = sender.getDestination(destinationId);
        if (serviceTypeDoesNotMatchDestination(destination, serviceType)) {
            throw new IllegalArgumentException("Provided service type '" + serviceType.getIdentifier() + "' is not allowed by the destination ");
        }
    }

    private void testDefaults(final SubmissionPayload submissionPayload) {

        if (submissionPayload.getDestinationId() == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (submissionPayload.getServiceType() == null) {
            throw new IllegalArgumentException("ServiceType is mandatory, but was null.");
        } else if (submissionPayload.getServiceType().getIdentifier() == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(submissionPayload.getServiceType().getIdentifier())) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        Destination destination = sender.getDestination(submissionPayload.getDestinationId());
        if (serviceTypeDoesNotMatchDestination(destination, submissionPayload.getServiceType())) {
            throw new IllegalArgumentException("Provided service type '" + submissionPayload.getServiceType().getIdentifier() + "' is not allowed by the destination ");
        }
        if (mimeTypeAndSchemaUriDoNotMatchDestination(destination, submissionPayload.getDataMimeType(), submissionPayload.getSchemaUri())) {
            throw new IllegalArgumentException("Combination of provided MIME type '" + submissionPayload.getDataMimeType() +
                    "' and schema URI '" + submissionPayload.getSchemaUri() + "' is not allowed by the destination");
        }
    }

    private boolean serviceTypeDoesNotMatchDestination(final Destination destination, final ServiceType serviceType) {
        return destination.getServices().stream()
                .map(DestinationService::getIdentifier)
                .filter(Objects::nonNull)
                .filter(serviceIdentifier -> serviceIdentifier.equals(serviceType.getIdentifier()))
                .findFirst()
                .isEmpty();
    }

    private boolean mimeTypeAndSchemaUriDoNotMatchDestination(final Destination destination, final MimeType mimeType, final URI schemaUri) {
        return destination.getServices().stream()
                .flatMap(service -> service.getSubmissionSchemas().stream())
                .filter(Objects::nonNull)
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .filter(schema -> schema.getSchemaUri().equals(schemaUri))
                .findFirst()
                .isEmpty();
    }

    private void testOnValidDataFormat(final SubmissionPayload submissionPayload) {
        final MimeType dataMimeType = submissionPayload.getDataMimeType();
        if (dataMimeType.equals(MimeType.APPLICATION_JSON)) {
            checkJsonFormat(submissionPayload);
        } else if (dataMimeType.equals(MimeType.APPLICATION_XML)) {
            checkXmlFormat(submissionPayload);
        }
    }

    private void checkXmlFormat(final SubmissionPayload submissionPayload) {
        final ValidationResult validationResult = sender.validateXmlFormat(submissionPayload.getData());
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected xml format, please provide valid xml: " + validationResult.getError().getMessage());
        }
    }

    private void checkJsonFormat(final SubmissionPayload submissionPayload) {
        final ValidationResult validationResult = sender.validateJsonFormat(submissionPayload.getData());
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected json format, please provide valid json: " + validationResult.getError().getMessage());
        }
    }

    private boolean noValidLeikaKeyPattern(final String leikaKey) {
        if (leikaKey.length() < 7 || leikaKey.length() > 255) {
            return true;
        }
        return !LEIKA_KEY_PATTERN.matcher(leikaKey).matches();
    }
}
